#!/usr/libexec/platform-python
#
# DISCLAIMER: this script is not supported by Red Hat.
# Usage: ./eval-ovl-size.py
# Give a rough estimate of the required available space to in-place upgrade.

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api

MIN_SIZE = 1500   # MB
EXTRA_MARGIN = 15 # %
import os, os.path, subprocess

def _shell(command):
    # works with python2.7 (el7) as well as python3 (el8)
    p = subprocess.Popen(command,
                     stdout=subprocess.PIPE,
                     stderr=subprocess.PIPE,
                     close_fds=True,
                     shell=True,
                     universal_newlines=True)
    output = []
    while True:
        line = p.stdout.readline().rstrip('\n')
        if line:
            output.append(line)
        else:
            break
    return output

def get_rpms_size():
    rpms_size = 0
    for size in _shell("rpm -qa --queryformat '%{SIZE}\n'"):
        rpms_size += int(size)
    return rpms_size

def get_xfs_ftype(mp):
    return int(_shell("xfs_info %s | grep -o ftype=." % mp)[0].split('=')[1])

def find_mnt(path):
    path = os.path.abspath(path)
    while not os.path.ismount(path):
        path = os.path.dirname(path)
    return path

def get_partitions():
    parts = []
    with open('/proc/mounts', 'r') as f:
        mounts = f.readlines()
    for mount in mounts:
        if not mount.startswith('/'):
            continue
        dev, mp, fstype, opts, _5, _6 = mount.split(' ')
        if mp.startswith('/run'):
            continue
        p = os.statvfs(mp)
        parts.append({
          'mp': mp,
          'fstype': fstype,
          'ftype': get_xfs_ftype(mp) if fstype == 'xfs' else -1,
          'avail': p.f_bavail * p.f_frsize
        })
    return parts

def iter_parts(lmp):
    avail = 0
    xfs_parts = 0
    ftype_case = 0
    for p in get_partitions():
        if p['mp'] == lmp:
            avail = int(p['avail'] / 1024**2)
        if p['fstype'] == 'xfs':
            xfs_parts += 1
        if p['ftype'] == 0:
            ftype_case += 1
    return avail, xfs_parts, ftype_case

def show_eval(lmp, avail, xfs_parts, ftype_case, rpms_size):
    summary = ("`/var/lib/bclinux-sysmt` resides on the `{0}` partition, Current available space in `{1}`: {2} MB. Installed size for all RPMs: {3} MB. "
               .format(lmp, lmp, avail, rpms_size))
    inhibitor = False
    hint = ''
    if xfs_parts and ftype_case:
        ovl_size = int(rpms_size + (rpms_size * EXTRA_MARGIN / 100))
        ovl_size = max(ovl_size, 2048) # don't use less than the default
        required = (ftype_case * ovl_size) + MIN_SIZE
        api.current_logger().info("== XFS ftype=0 case ==")
        summary += "XFS ftype=0 case, "
        env_size = int(os.getenv('LEAPP_OVL_SIZE', default='2048'))
        if ovl_size > env_size:
            hint = ("Adjust $LEAPP_OVL_SIZE envionment variables before running Leapp.")
            commands = [['echo', '"export', 'LEAPP_OVL_SIZE=%s"' % ovl_size, '>>', '/root/.bashrc', '&&', 'source', '/root/.bashrc']]
            inhibitor = True
        summary += "Required space in the `%s` partition is: NumberOfXfsMountedPartitions * LEAPP_OVL_SIZE = %s MB. " % (lmp, required)
        if avail < required:
            estimated = required - avail + MIN_SIZE
            summary += "There is not enough space in `%s`, extend it at least by %s MB. " % (lmp, estimated)
            inhibitor = True
        summary += """If Leapp/DNF requires again more space: adjust $LEAPP_OVL_SIZE with a bigger size than %s(echo "export LEAPP_OVL_SIZE=xxxx" >> /root/.bashrc && source /root/.bashrc), AND extend the `%s` partition accordingly. """ % (ovl_size, lmp)
    else:
        required = int(rpms_size + (rpms_size * EXTRA_MARGIN / 100) + MIN_SIZE)
        api.current_logger().info("== Standard case ==")
        summary += "Standard case, "
        if avail < required:
            estimated = required - avail + MIN_SIZE
            summary += "There is not enough space in `%s`, extend it at least by %s MB. " % (lmp, estimated)
            inhibitor = True
        summary += "If Leapp/DNF requires again more space, you need to extend the `%s` partition. " % lmp

    summary += "For more information, please read https://access.redhat.com/solutions/5057391"

    report = [
        reporting.Title("whether lack of the free space on the disk."),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Tags([reporting.Tags.SANITY]),
    ]
    if inhibitor:
        report.append(reporting.Flags([reporting.Flags.INHIBITOR]))
        if hint:
            report.append(reporting.Remediation(hint=hint, commands=commands))
    reporting.create_report(report)

def process():
    lmp = find_mnt('/var/lib/bclinux-sysmt')
    avail, xfs_parts, ftype_case = iter_parts(lmp)
    rpms_size = int(get_rpms_size() / 1024**2)
    show_eval(lmp, avail, xfs_parts, ftype_case, rpms_size)
