import shutil

from leapp.actors import Actor
from leapp.libraries.common import dnfplugin
from leapp.libraries.stdlib import run
from leapp.models import (
    DNFPluginTask,
    FilteredRpmTransactionTasks,
    RHSMInfo,
    StorageInfo,
    TargetUserSpaceInfo,
    TransactionCompleted,
    UsedTargetRepositories,
)
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag
import subprocess
import os


class DnfUpgradeTransaction(Actor):
    """
    Setup and call DNF upgrade command

    Based on previously calculated RPM transaction data, this actor will setup and call
    rhel-upgrade DNF plugin with necessary parameters
    """

    name = 'dnf_upgrade_transaction'
    consumes = (
        DNFPluginTask,
        FilteredRpmTransactionTasks,
        RHSMInfo,
        StorageInfo,
        TargetUserSpaceInfo,
        UsedTargetRepositories,
    )
    produces = (TransactionCompleted,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    migrear_exe_file = '/usr/sbin/migrear'
    migrear_param_file = '/etc/rear/grub_param.json'

    def process(self):
        src_rhsm_info = next(self.consume(RHSMInfo), None)
        if src_rhsm_info:
            for prod_cert in src_rhsm_info.existing_product_certificates:
                run(['rm', '-f', prod_cert])

        used_repos = self.consume(UsedTargetRepositories)
        storage_info = next(self.consume(StorageInfo), None)
        plugin_info = list(self.consume(DNFPluginTask))
        tasks = next(self.consume(FilteredRpmTransactionTasks), FilteredRpmTransactionTasks())
        target_userspace_info = next(self.consume(TargetUserSpaceInfo), None)

        try:
            dnfplugin.perform_transaction_install(
                tasks=tasks, used_repos=used_repos, storage_info=storage_info, target_userspace_info=target_userspace_info,
                plugin_info=plugin_info
            )
        except Exception as e:
            raise e
        finally:
            if os.path.isfile(self.migrear_exe_file) and os.path.isfile(self.migrear_param_file):
                try:
                    subprocess.check_call("migrear --recover_preset no", shell=True)
                except Exception as e:
                    self.log.error(" ".join(["Failed restore migrear entry", "Exception:", str(e)]), exc_info=True)
            else:
                self.log.warning("Failed restore migrear entry. Executive or parameter file Not exist!", exc_info=True)

        self.produce(TransactionCompleted())
        userspace = next(self.consume(TargetUserSpaceInfo), None)
        if userspace:
            try:
                shutil.rmtree(userspace.path)
            except Exception as e:
                self.log.warning("Failed to remove temporary userspace - error ignored. " + repr(e), exc_info=True)
                try:
                    subprocess.check_call("rm -fr " + userspace.path, shell=True)
                except Exception as e:
                    self.log.warning("Failed to delete temporary userspace - error ignored", exc_info=True)
