import os
import subprocess
import re
import json
import time
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import run, CalledProcessError, api
from leapp.models import EnvVar, OSRelease

CURRENT_TARGET_VERSION = '8.2'
CUSTOM_REPO_PATH = "/etc/bclinux-sysmt/files/bclinux-sysmt_upgrade_repositories.repo"

ENV_IGNORE = ('LEAPP_CURRENT_PHASE', 'LEAPP_CURRENT_ACTOR', 'LEAPP_VERBOSE',
              'LEAPP_DEBUG')

ENV_MAPPING = {'LEAPP_DEVEL_DM_DISABLE_UDEV': 'DM_DISABLE_UDEV'}


LEAPP_UPGRADE_FLAVOUR_DEFAULT = 'default'
LEAPP_UPGRADE_FLAVOUR_SAP_HANA = 'saphana'

HANA_BASE_PATH = '/hana/shared'
HANA_SAPCONTROL_PATH = 'exe/linuxx86_64/hdb/sapcontrol'
TARGET_SCRIPT_PATH = '/usr/local/sysom/node/customized/secure_scripts/'
MIGRATION_SCRIPT_PATH = '/var/lib/bclinux-sysmt/customized/scripts/'

def get_env_vars():
    """
    Gather LEAPP_DEVEL environment variables and respective mappings to provide them as messages to be
    available after reboot.
    """
    env_vars = []
    leapp_vars = {k: v for (k, v) in os.environ.items() if k.startswith('LEAPP_') and k not in ENV_IGNORE}
    for k, v in leapp_vars.items():
        if k in ENV_MAPPING:
            env_vars.append(EnvVar(name=ENV_MAPPING.get(k), value=v))
            continue
        env_vars.append(EnvVar(name=k, value=v))

    return env_vars

def get_os_version_id(version_id):
    if re.match('\d\.\d', version_id):
        return version_id
    else:
        """CentOS version_id is always 7 or 8, so we retrive from centos-release."""

        old_version = subprocess.check_output("rpm -q --whatprovides /etc/redhat-release", shell=True)
        old_version = str(old_version).encode("utf-8")
        old_version = old_version.split('\n')[:-1]
        old_version = old_version[0]
        if (re.match('centos-linux-release', old_version)) or (re.match('bclinux-release-ldk', old_version)):
            mainver = old_version.split('-')[3]
            subver = old_version.split('-')[4].split('.')[0]
            return mainver+'.'+subver
        elif (re.match('centos-release', old_version)) or (re.match('bclinux-release', old_version)):
            mainver = old_version.split('-')[2]
            subver = old_version.split('-')[3].split('.')[0]
            return mainver+'.'+subver
        else:
            raise IOError('The version_id of OS is wrong format.')

def get_os_release(path):
    """Retrieve data about System OS release from provided file."""
    try:
        with open(path) as f:
            data = dict(l.strip().split('=', 1) for l in f.readlines() if '=' in l)
            return OSRelease(
                release_id=data.get('ID', '').strip('"'),
                name=data.get('NAME', '').strip('"'),
                pretty_name=data.get('PRETTY_NAME', '').strip('"'),
                version=data.get('VERSION', '').strip('"'),
                version_id=get_os_version_id(data.get('VERSION_ID', '').strip('"')),
                variant=data.get('VARIANT', '').strip('"') or None,
                variant_id=data.get('VARIANT_ID', '').strip('"') or None
            )
    except IOError as e:
        raise StopActorExecutionError(
            message='Cannot collect the system OS facts.',
            details={'details': str(e)}
        )


def get_booted_kernel():
    """Get version and release of the currently used kernel in one string."""
    try:
        return run(['/usr/bin/uname', '-r'])['stdout'].strip()
    except CalledProcessError as e:
        raise StopActorExecutionError(
            message='Unable to obtain release of the booted kernel.',
            details={'details': str(e), 'stderr': e.stderr}
        )


def get_target_version(flavour=LEAPP_UPGRADE_FLAVOUR_DEFAULT):
    """
    Get migration target OS version from CUSTOM_REPO_PATH file, since 'leapp customrepo --seturl' command could set this file.
    """
    current_target_version = CURRENT_TARGET_VERSION
    command = "grep -m 1 'baseurl' " + CUSTOM_REPO_PATH + " | grep -o 'el8.[1-9]'"
    try:
        result = subprocess.check_output(command, shell=True)
        current_target_version = result[2:5]
    except Exception as e:
        api.current_logger().warning('Command [{0}] failed to get migration target OS version. Please Check!'.format(command))

    return current_target_version


def detect_sap_hana():
    """ Detect SAP HANA based on existance of /hana/shared/*/exe/linuxx86_64/hdb/sapcontrol """
    if os.path.exists(HANA_BASE_PATH):
        for entry in os.listdir(HANA_BASE_PATH):
            # Does /hana/shared/{entry}/exe/linuxx86_64/hdb/sapcontrol exist?
            if os.path.exists(os.path.join(HANA_BASE_PATH, entry, HANA_SAPCONTROL_PATH)):
                return True
    return False


def get_upgrade_flavour():
    """ Returns the flavour of the upgrade for this system."""
    if detect_sap_hana():
        return LEAPP_UPGRADE_FLAVOUR_SAP_HANA
    return LEAPP_UPGRADE_FLAVOUR_DEFAULT

def origin_script_changed(origin_file, target_file):
    if os.path.isfile(target_file):
        o_mtime = os.path.getmtime(origin_file)
        t_mtime = os.path.getmtime(target_file)
        return o_mtime > t_mtime
    else:
        return True

def prepare_process_script():
    migration_prescript_origin = TARGET_SCRIPT_PATH + '01_upload_script.sh'
    migration_postscript_origin = TARGET_SCRIPT_PATH + '02_upload_script.sh'
    migration_prescript = MIGRATION_SCRIPT_PATH + "prev/" + "10_migration_prescript"
    migration_postscript = MIGRATION_SCRIPT_PATH + "post/" + "10_migration_postscript"

    try:
        if os.path.isfile(migration_prescript_origin):
            if origin_script_changed(migration_prescript_origin, migration_prescript):
                subprocess.check_call("/bin/cp {0} {1}".format(migration_prescript_origin, migration_prescript), shell=True)
                subprocess.check_call("chmod +x {0}".format(migration_prescript), shell=True)
                api.current_logger().info('migration pre_process script {} Loaded.'.format(migration_prescript))
        else:
            subprocess.check_call("rm -f " + migration_prescript, shell=True)

        if os.path.isfile(migration_postscript_origin):
            if origin_script_changed(migration_postscript_origin, migration_postscript):
                subprocess.check_call("/bin/cp {0} {1}".format(migration_postscript_origin, migration_postscript), shell=True)
                subprocess.check_call("chmod +x {0}".format(migration_postscript), shell=True)
                api.current_logger().info('migration post_process script {} Loaded.'.format(migration_postscript))
        else:
            subprocess.check_call("rm -f " + migration_postscript, shell=True)
    except Exception as e:
        api.current_logger().warning('Prepare process script Failed! Please Check! {}'.format(e))

def execute_preprocess_script():
    migration_preprocess_path = MIGRATION_SCRIPT_PATH + "prev"
    script_dir_list = os.listdir(migration_preprocess_path)
    script_file_list = []
    for dir_item in script_dir_list:
        dir_item_path = migration_preprocess_path + '/' + dir_item
        if os.path.isfile(dir_item_path) and os.access(dir_item_path, os.X_OK):
            script_file_list.append(dir_item)

    if script_file_list:
        script_file_list.sort()
        api.current_logger().info('Now execute the prev-process scripts in order, [{}].'.format(script_file_list))

        script_output_path = '/var/log/bclinux-sysmt/scripts_output'
        subprocess.check_call("mkdir -p {0}".format(script_output_path), shell=True)
        for script_exec in script_file_list:
            script_exec_path = migration_preprocess_path + '/' + script_exec
            output_file = script_output_path + "/prev_proc_" + time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime()) + ".out"
            r = run([script_exec_path])
            api.current_logger().info('Script: {0}, exit_code: {1}, pid: {2}, signal: {3}.'.format(script_exec, r["exit_code"], r["pid"], r["signal"]))

            with open(output_file, 'w') as out_fd:
                out_fd.write("Script {} stderr:\n".format(script_exec))
                out_fd.write(r["stderr"])
                out_fd.write('\n')
                out_fd.write("Script {} stdout:\n".format(script_exec))
                out_fd.write(r["stdout"])
            if r['exit_code'] == 0:
                run(['chmod', '644', script_exec_path])

def preprocess():
    prepare_process_script()
    execute_preprocess_script()
